# Copyright 2018 Immuta, Inc. Licensed under the Immuta Software License
# Version 0.1 (the "License"); you may not use this file except in
# compliance with the License. You may obtain a copy of the License at
#
#    http://www.immuta.com/licenses/Immuta_Software_License_0.1.txt
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.


class InvalidTokenError(Exception):
    def __init__(self, msg):
        super(InvalidTokenError, self).__init__(msg)
        self.message = msg


class InvalidLoginError(Exception):
    def __init__(self, msg):
        super(InvalidLoginError, self).__init__(msg)
        self.message = msg


class UnknownAuthenticationScheme(Exception):
    def __init__(self, msg=None):
        msg = (
            msg
            if msg
            else "Unable to generate authentication scheme with provided parameters."
        )
        super(UnknownAuthenticationScheme, self).__init__(msg)
        self.message = msg


class ImmutaCredentialsError(Exception):
    pass


class BadImmutaConfigException(Exception):
    pass



# Copyright 2018 Immuta, Inc. Licensed under the Immuta Software License
# Version 0.1 (the "License"); you may not use this file except in
# compliance with the License. You may obtain a copy of the License at
#
#    http://www.immuta.com/licenses/Immuta_Software_License_0.1.txt
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
from typing import Optional, List, Dict, Any, Union, Sequence, Iterator

import requests
import six.moves.urllib_parse as urlparse
from requests import HTTPError
from requests.packages.urllib3.exceptions import InsecureRequestWarning
from urllib3.util.retry import Retry

from .authenticate import ImmutaRequestsAuth, build_auth_scheme, retrieve_credentials
from .data_source import (
    DataSource,
    DataSourceColumn,
    DataSourceDictionary,
    Handler,
    blob_handler_type,
)
from .policy import GlobalPolicy, make_policy_object_from_json
from .log import LoggingMixin

requests.packages.urllib3.disable_warnings(InsecureRequestWarning)

MAX_RETRIES = 5


class ImmutaSession(requests.Session):
    def __init__(self, immuta_url, auth_scheme=None, ca_certs=True):
        super(ImmutaSession, self).__init__()
        if immuta_url[-1] != "/":
            immuta_url = immuta_url + "/"
        self.immuta_url = immuta_url
        self.auth = ImmutaRequestsAuth(immuta_url, auth_scheme, ca_certs)
        self.verify = ca_certs
        retries = Retry(
            total=MAX_RETRIES, backoff_factor=0.1, status_forcelist=[500, 502, 503, 504]
        )
        self.mount("https://", requests.adapters.HTTPAdapter(max_retries=retries))

    def request(self, method, url, *args, **kwargs):
        url = urlparse.urljoin(self.immuta_url, url)
        kwargs["verify"] = self.verify
        return super(ImmutaSession, self).request(method, url, *args, **kwargs)


class ImmutaClient(LoggingMixin):
    def __init__(
        self, base_url=None, auth_scheme=None, ca_certs=True, session=None, **kwargs
    ):
        if session:
            self._session = session
            self.base_url = session.immuta_url
        else:
            if not auth_scheme:
                auth_scheme = build_auth_scheme(**kwargs)
            self._session = ImmutaSession(base_url, auth_scheme, ca_certs)
            self.base_url = base_url

    def __remove_blob_handler_attributes(self, request_prefix, blob_handler):
        allowed = None
        if request_prefix == "elastic":
            allowed = [
                "bodataTableName",
                "dataSourceName",
                "masked",
                "eventTimeAttribute",
                "policyAttributes",
                "staleDataTolerance",
                "elasticProperties",
                "directoryStructure",
                "query",
            ]
        elif request_prefix in ["s3", "hdfs"]:
            allowed = ["ingestUserId", "ingestAPIKey"]
        if allowed:
            for key in blob_handler["metadata"].keys():
                if key not in allowed:
                    del blob_handler["metadata"][key]

    def _buildVisibilitySchema(self, policy_handler):
        json_policies = policy_handler["jsonPolicies"]
        fields = set()
        for policy in json_policies:
            if policy["type"] == "rowOrObjectRestriction":
                for rule in policy["rules"]:
                    for condition in rule["config"]["qualifications"]["conditions"]:
                        fields.add(condition["field"])
        return {"fields": list(fields)}

    @classmethod
    def make_glob_request_headers(cls, config: Dict[str, Any]) -> Dict[str, Any]:
        if config["handler_type"] in ["PostgreSQL", "Redshift"]:
            return cls.make_generic_odbc_request_headers(config)
        if config["handler_type"] == "Amazon Athena":
            return cls.make_athena_glob_request_headers(config)
        raise TypeError

    @classmethod
    def make_athena_glob_request_headers(cls, config: Dict[str, Any]) -> Dict[str, str]:
        return {
            "sql-authentication-type": "accessKey",
            "sql-aws-region": config["region"],
            "sql-aws-result-location": config["queryResultLocationBucket"],
            "sql-ssl": "true",
            "sql-username": config["username"],
            "sql-password": config["password"],
        }

    @classmethod
    def make_generic_odbc_request_headers(
        cls, config: Dict[str, Any]
    ) -> Dict[str, Any]:
        return {
            "sql-hostname": config["hostname"],
            "sql-port": f"{config['port']}",
            "sql-ssl": "true",
            "sql-username": config["username"],
            "sql-password": config["password"],
        }

    @classmethod
    def make_get_request_params(
        cls,
        search_text: Optional[str] = None,
        public_only: Optional[bool] = None,
        name_only: Optional[bool] = False,
        mode: Optional[int] = 0,
        size: Optional[int] = 50,
        offset: Optional[int] = 0,
    ) -> Dict[str, Any]:
        params = {
            "searchText": search_text,
            "publicOnly": public_only,
            "nameOnly": name_only,
            "mode": mode,
            "size": size,
            "offset": offset,
        }
        clean_params = {k: v for k, v in params.items() if v is not None}
        return clean_params

    def get(self, path, *args, **kwargs):
        resp = self._session.get(path, *args, **kwargs)
        if resp.status_code != 200:
            self.log.error(
                f"Error in request. Response status: {resp.status_code}, text:"
                f" {resp.text}"
            )
        resp.raise_for_status()
        return resp.json()

    def post(self, path, data, *args, **kwargs):
        resp = self._session.post(path, json=data, *args, **kwargs)
        resp.raise_for_status()
        return resp

    def put(self, path, data, *args, **kwargs):
        resp = self._session.put(path, json=data, *args, **kwargs)
        resp.raise_for_status()
        return resp

    def delete(self, path, *args, **kwargs):
        resp = self._session.delete(path, *args, **kwargs)
        resp.raise_for_status()
        return resp

    def get_api_key(self):
        endpoint = "bim/apikey"
        res = self._session.post(endpoint)
        res.raise_for_status()
        return res.json().get("apikey", None)

    def revoke_api_key(self, keyid) -> bool:
        endpoint = "bim/apikey/{}".format(keyid)
        res = self._session.delete(endpoint)
        res.raise_for_status()
        return True

    def create_tag(
        self, tag_data: Dict[str, Any], raise_on_existing_tag: bool = False
    ) -> bool:
        res = self._session.post("tag", json=tag_data)
        if res.status_code == 400:
            if (
                "overlap with existing hierarchies" in res.json()["message"]
                and not raise_on_existing_tag
            ):
                return True
        res.raise_for_status()
        return True

    def get_table_names(self, config: Dict[str, Any]) -> List[Dict[str, str]]:
        """
        Queries the given database and returns a list of all tables with their schema names.
        NOTE: This is an undocumented API endpoint that might disappear in a future release of Immuta!
        """
        headers = self.make_glob_request_headers(config)
        return self.get(
            f"{blob_handler_type(config['handler_type'])}/database/{config['database']}/table",
            headers=headers,
        )

    def get_column_types(
        self, data_source_type: str, handler: Handler, config: Dict[str, Any]
    ) -> List[DataSourceColumn]:
        """
        Returns info on all columns that exist in the external data source specified via the handler.
        NOTE: This is an undocumented API endpoint that might disappear in a future release of Immuta!
        """
        headers = self.make_glob_request_headers(config)
        data = handler.dict(
            by_alias=True,
            include={"metadata": {"database", "table", "fh_schema", "format"}},
        )
        res = self.post(
            f"{blob_handler_type(data_source_type)}/handler/getColumnTypes",
            headers=headers,
            data=data,
        )
        res.raise_for_status()
        return [DataSourceColumn(**c) for c in res.json()]

    def get_data_source_dictionary(self, id: int) -> DataSourceDictionary:
        res = self._session.get(f"dictionary/{id}")
        res.raise_for_status()
        return DataSourceDictionary(**res.json())

    def update_data_source_dictionary(
        self, id: int, dictionary: DataSourceDictionary
    ) -> bool:
        res = self._session.put(
            f"dictionary/{id}", data=dictionary.json(skip_defaults=True)
        )
        res.raise_for_status()
        return True

    def create_data_sources(self, handler_type, handlers, data_source):
        post_body = {"handler": handlers, "dataSource": data_source}
        url = "{}/handler".format(handler_type)
        handler_response = self._session.post(url, json=post_body)
        if handler_response.status_code == 200:
            self.log.debug("Bulk data source create job submitted successfully")
            return True
        self.log.error("Error: %s", handler_response.text)
        handler_response.raise_for_status()

    def create_data_source(
        self,
        data_source: DataSource,
        handler: Union[Handler, Sequence[Handler]],
        policy_handler=None,
        handler_base_url=None,
        dictionary=None,
    ):
        request_prefix = blob_handler_type(data_source.blobHandlerType)
        is_bulk_insert = isinstance(handler, list)
        if is_bulk_insert:
            handlers = [h.dict(by_alias=True, skip_defaults=True) for h in handler]
        elif isinstance(handler, Handler):  # type check here to make mypy happy
            handlers = handler.dict(by_alias=True, skip_defaults=True)
        else:
            raise RuntimeError(
                "Invalid format for given blob handler. Expected either a list or a"
                " BlobHandler object"
            )
        post_body = {
            "handler": handlers,
            "dataSource": data_source.dict(by_alias=True, skip_defaults=True),
        }
        if policy_handler:
            post_body["policyRules"] = policy_handler["jsonPolicies"]
        self.log.debug(post_body)
        handler_response = self._session.post(
            f"{request_prefix}/handler", json=post_body
        )
        self.log.debug("Response: %s", handler_response.text)
        # There's no ID given back for bulk create requests
        if handler_response.status_code == 200:
            if not is_bulk_insert:
                return self.get_data_source(handler_response.json()["dataSourceId"])
            return None
        if "already exists" in handler_response.text:
            self.log.info("Data source with name %s already exists", data_source.name)
            return None
        self.log.error("Error: %s", handler_response.text)
        handler_response.raise_for_status()

    def update_data_source(
        self,
        data_source: DataSource,
        handler: Handler = None,
        policy_handler=None,
        handler_base_url=None,
        dictionary=None,
    ):

        request_prefix = blob_handler_type(data_source.blobHandlerType)

        if policy_handler:
            visibility_schema = self._buildVisibilitySchema(policy_handler)
            policy_handler["dataSourcePolicyHandler"] = data_source.get("policyHandler")
            if (data_source.get("policyHandler") or {}).get("handlerId"):
                handler_id = data_source["policyHandler"]["handlerId"]
                access_key = None
                handler_url = f"{handler_base_url}/policy/handler/{handler_id}"
                resp = self._session.put(handler_url, json=policy_handler)
                resp.raise_for_status()
            else:
                policy_handler = self.create_policy_handler(policy_handler)
                handler_id = policy_handler["id"]
                handler_url = (
                    f"{handler_base_url}/policy/handler/{policy_handler['id']}"
                )
                access_key = policy_handler["accessKey"]

            data_source["policyHandler"] = {
                "url": handler_url,
                "handlerId": handler_id,
                "visibilitySchema": visibility_schema,
            }
            if access_key:
                data_source["policyHandler"]["accessKey"] = access_key

        # update the data source
        resp = self._session.put(f"dataSource/{data_source['id']}", json=data_source)
        resp.raise_for_status()
        saved_data_source = resp.json()

        # update the dictionary if it was passed in
        if dictionary:
            resp = self._session.put(f"dictionary/{data_source['id']}", json=dictionary)
            resp.raise_for_status()

        # update the blob handler if it was passed in
        if handler:
            blob_handler_id = data_source["blobHandler"]["url"].split("/")[-1]
            self.__remove_blob_handler_attributes(
                blob_handler_type(data_source), handler
            )
            handler_url = (
                f"{handler_base_url}/{request_prefix}/handler/{blob_handler_id}"
            )
            resp = self._session.put(handler_url, json=handler)
            resp.raise_for_status()

        return saved_data_source

    def get_data_source(self, id=None, update_stats=False, name=None):
        if not id:
            if not name:
                raise Exception("Either id or name must be provided but were not.")
            return self.get_data_source_by_name(name)

        endpoint = f"dataSource/{id}"
        return self.get(endpoint, params={"updateStats": str(update_stats).lower()})

    def get_data_source_list(
        self,
        search_text=None,
        public_only=None,
        name_only=False,
        mode=0,
        size=50,
        offset=0,
    ):
        params = self.make_get_request_params(
            search_text=search_text,
            public_only=public_only,
            name_only=name_only,
            mode=mode,
            size=size,
            offset=offset,
        )
        return self.get("dataSource", params=params)

    def get_data_source_by_name(self, name):
        endpoint = "dataSource/name/{}".format(name)
        return self.get(endpoint)

    def get_global_policies(
        self,
        search_text: Optional[str] = None,
        public_only: Optional[bool] = None,
        mode: Optional[int] = None,
        name_only: Optional[bool] = False,
    ) -> Iterator[GlobalPolicy]:
        params = self.make_get_request_params(
            search_text=search_text,
            public_only=public_only,
            name_only=name_only,
            mode=mode,
            size=None,
            offset=None,
        )
        res = self.get("policy/global", params=params)
        for policy in res:
            yield (make_policy_object_from_json(policy))

    def create_policy_handler(self, handler: Dict[str, str]) -> Dict[str, str]:
        handler_response = self._session.post("policy/handler", json=handler)
        handler_response.raise_for_status()
        return handler_response.json()

    def create_global_policy(self, policy: GlobalPolicy) -> Union[int, bool]:
        res = self._session.post("policy/global", data=policy.json(by_alias=True))
        if res.status_code == 200:
            return res.json()["id"]
        if res.status_code == 422 and res.json()["validation"][0]["code"] == "unique":
            self.log.warn(f"Policy with name {policy.name} already exists.")
        res.raise_for_status()
        return False

    def update_global_policy(
        self, policy: GlobalPolicy, id: Optional[int]
    ) -> Dict[str, Any]:
        res = self._session.put(f"policy/global/{id}", data=policy.json(by_alias=True))
        res.raise_for_status()
        return res.json()

    def delete_global_policy(self, id: Optional[int]) -> bool:
        res = self.delete(f"policy/global/{id}", params={"policyId": id})
        res.raise_for_status()
        return True

    def disable_data_source(
        self, id: Optional[int] = None, name: Optional[str] = None
    ) -> int:
        """
        Disable a data source with the id or the name provided.
        If both id and name are provided, the name will be ignored
        :param id: data source Id
        :param name: data source name
        :return: the data source id
        """
        data_source = self.get_data_source(id=id, name=name)
        if data_source["deleted"]:
            self.log.warning(
                f"Data source \"{data_source['name']}\" id: {data_source['id']} is"
                " disabled already"
            )
        else:
            self.delete(f"dataSource/{data_source['id']}")
        return data_source["id"]

    def restore_data_source(
        self, id: Optional[int] = None, name: Optional[str] = None
    ) -> int:
        """
        Restore a data source that was previously disabled with the id or the name provided.
        If both id and name are provided, the name will be ignored
        :param id: data source Id
        :param name: data source name
        :return: the data source id
        """
        data_source = self.get_data_source(id=id, name=name)
        if not data_source["deleted"]:
            self.log.warning(
                f"Data source \"{data_source['name']}\" id: {data_source['id']} is"
                " enabled already"
            )
        else:
            self.put(f"dataSource/{data_source['id']}", data={"deleted": False})
        return data_source["id"]

    def delete_data_source(
        self, id: Optional[int] = None, name: Optional[str] = None
    ) -> int:
        """
        Deletes completely a data source with the id or the name provided.
        If both id and name are provided, the name will be ignored
        :param id: data source Id
        :param name: data source name
        :return: the data source id
        """
        if not id and not name:
            raise Exception("Either id or name must be provided but were not.")
        try:
            if not id and name:
                id = self.get_data_source_by_name(name=name)["id"]

            # To completely remove a data source, we need to disabled it first,
            # but Immuta uses the same endpoint for both actions,
            # Due to this reason, we run `self.delete()` twice
            # unless the data source was already disabled
            result = self.delete(f"dataSource/{id}")

            if not result.json()["hardDelete"]:
                # if the data source was only disabled, delete it again
                self.delete(f"dataSource/{id}")
        except HTTPError as e:
            if e.response.status_code == 404:
                self.log.warning(
                    f"Data source {id or name} not found, it might have been deleted"
                    " already"
                )
            else:
                raise

        assert isinstance(id, int)
        return id

    def tag_data_source(self, id: int, tag_data: List[Dict[str, Any]]) -> bool:
        """
        Adds tags to the data source directly, not to columns within the data source.
        :param id: data source id
        :param tag_data: list of tag dicts to apply to the data source
        :return: True
        """
        res = self._session.post(f"tag/datasource/{id}", json=tag_data)
        res.raise_for_status()
        return True


def get_client(base_url: str, auth_config: Dict[str, Any], **kwargs) -> ImmutaClient:
    return ImmutaClient(base_url=f"https://{base_url}", **auth_config)



from collections import namedtuple
from typing import Dict, List, Any
from pydantic import ValidationError

import pytest

import fh_immuta_utils.data_source as ds

NameTestKeys = namedtuple(
    "NameTestKeys", ["handler_type", "schema", "table", "user_prefix", "expected_name"]
)

IMMUTA_NAME_TESTS = [
    NameTestKeys(
        handler_type="PostgreSQL",
        schema="foo",
        table="barbazquxquux",
        user_prefix="",
        expected_name=f"{ds.PREFIX_MAP['PostgreSQL']}_foo_barbazquxquux",
    ),
    NameTestKeys(
        handler_type="PostgreSQL",
        schema="foo",
        table="barbazquxquux",
        user_prefix="quuz",
        expected_name=f"quuz_{ds.PREFIX_MAP['PostgreSQL']}_foo_barbazquxquux",
    ),
    NameTestKeys(
        handler_type="PostgreSQL",
        schema="foo",
        table=f"{'a'*(ds.MAX_IMMUTA_NAME_LIMIT - 7)}",
        user_prefix="",
        expected_name=(
            f"{ds.PREFIX_MAP['PostgreSQL']}_foo_{'a'*(ds.MAX_IMMUTA_NAME_LIMIT - 7)}"
        ),
    ),
    NameTestKeys(
        handler_type="PostgreSQL",
        schema="foo",
        table=f"{'a'*ds.MAX_IMMUTA_NAME_LIMIT}",
        user_prefix="",
        expected_name=(
            f"{ds.PREFIX_MAP['PostgreSQL']}_foo_{'a'*(ds.MAX_IMMUTA_NAME_LIMIT - 7)}"
        ),
    ),
    NameTestKeys(
        handler_type="Redshift",
        schema="foo",
        table="barbazquxquux",
        user_prefix="",
        expected_name=f"{ds.PREFIX_MAP['Redshift']}_foo_barbazquxquux",
    ),
    NameTestKeys(
        handler_type="Amazon S3",
        schema="foo",
        table="barbazquxquux",
        user_prefix="",
        expected_name=f"{ds.PREFIX_MAP['Amazon S3']}_foo_barbazquxquux",
    ),
    NameTestKeys(
        handler_type="Amazon Athena",
        schema="foo",
        table="barbazquxquux",
        user_prefix="",
        expected_name=f"{ds.PREFIX_MAP['Amazon Athena']}_foo_barbazquxquux",
    ),
]

POSTGRES_NAME_TESTS = [
    NameTestKeys(
        handler_type="PostgreSQL",
        schema="foo",
        table="barbazquxquux",
        user_prefix="",
        expected_name=f"{ds.PREFIX_MAP['PostgreSQL']}_foo_barbazquxquux",
    ),
    NameTestKeys(
        handler_type="PostgreSQL",
        schema="foo",
        table="barbazquxquux",
        user_prefix="quuz",
        expected_name=f"quuz_{ds.PREFIX_MAP['PostgreSQL']}_foo_barbazquxquux",
    ),
    NameTestKeys(
        handler_type="PostgreSQL",
        schema="foo",
        table=f"{'a'*(ds.MAX_POSTGRES_NAME_LIMIT - 7)}",
        user_prefix="",
        expected_name=(
            f"{ds.PREFIX_MAP['PostgreSQL']}_foo_{'a'*(ds.MAX_POSTGRES_NAME_LIMIT - 7)}"
        ),
    ),
    NameTestKeys(
        handler_type="PostgreSQL",
        schema="foo",
        table=f"{'a'*ds.MAX_POSTGRES_NAME_LIMIT}",
        user_prefix="",
        expected_name=(
            f"{ds.PREFIX_MAP['PostgreSQL']}_foo_{'a'*(ds.MAX_POSTGRES_NAME_LIMIT - 7)}"
        ),
    ),
    NameTestKeys(
        handler_type="Redshift",
        schema="foo",
        table="barbazquxquux",
        user_prefix="",
        expected_name=f"{ds.PREFIX_MAP['Redshift']}_foo_barbazquxquux",
    ),
    NameTestKeys(
        handler_type="Amazon S3",
        schema="foo",
        table="barbazquxquux",
        user_prefix="",
        expected_name=f"{ds.PREFIX_MAP['Amazon S3']}_foo_barbazquxquux",
    ),
    NameTestKeys(
        handler_type="Amazon Athena",
        schema="foo",
        table="barbazquxquux",
        user_prefix="",
        expected_name=f"{ds.PREFIX_MAP['Amazon Athena']}_foo_barbazquxquux",
    ),
]


@pytest.mark.parametrize(
    "handler_type,schema,table,user_prefix,expected_name", IMMUTA_NAME_TESTS
)
def test_make_immuta_table_name(
    handler_type: str, schema: str, table: str, user_prefix: str, expected_name: str
):
    name = ds.make_immuta_table_name(
        handler_type=handler_type, schema=schema, table=table, user_prefix=user_prefix
    )
    assert len(name) <= ds.MAX_IMMUTA_NAME_LIMIT
    assert (
        name[: ds.MAX_IMMUTA_NAME_LIMIT - 8]
        == expected_name[: ds.MAX_IMMUTA_NAME_LIMIT - 8]
    )


@pytest.mark.parametrize(
    "handler_type,schema,table,user_prefix,expected_name", POSTGRES_NAME_TESTS
)
def test_make_postgres_table_name(
    handler_type: str, schema: str, table: str, user_prefix: str, expected_name: str
):
    name = ds.make_postgres_table_name(
        handler_type=handler_type, schema=schema, table=table, user_prefix=user_prefix
    )
    assert len(name) <= ds.MAX_POSTGRES_NAME_LIMIT
    assert name == expected_name[: ds.MAX_POSTGRES_NAME_LIMIT]


MetadataTestKeys = namedtuple(
    "MetadataTestKeys", ["db_keys", "handler_type", "expected_type", "kwargs"]
)
HANDLER_METADATA_TESTS = [
    MetadataTestKeys(
        db_keys={"hostname": "qux"},
        handler_type="PostgreSQL",
        expected_type=ds.PostgresHandlerMetadata,
        kwargs={},
    ),
    MetadataTestKeys(
        db_keys={"hostname": "qux"},
        handler_type="Redshift",
        expected_type=ds.PostgresHandlerMetadata,
        kwargs={},
    ),
    MetadataTestKeys(
        db_keys={
            "queryResultLocationBucket": "qux",
            "queryResultLocationDirectory": "quux",
            "region": "quz",
        },
        handler_type="Amazon Athena",
        expected_type=ds.AthenaHandlerMetadata,
        kwargs={},
    ),
    MetadataTestKeys(
        db_keys={"hostname": "qux"},
        handler_type="PostgreSQL",
        expected_type=ds.PostgresHandlerMetadata,
        kwargs={"bodataTableName": "foobar", "dataSourceName": "bazqux"},
    ),
]


@pytest.mark.parametrize(
    "db_keys,handler_type,expected_type, kwargs", HANDLER_METADATA_TESTS
)
def test_make_handler_metadata(
    db_keys: Dict[str, str],
    handler_type: str,
    expected_type: Any,
    kwargs: Dict[str, str],
):
    base_config = {"username": "foo", "password": "bar", "database": "baz"}

    config = {**base_config, **db_keys}
    with pytest.raises(KeyError):
        ds.make_handler_metadata(table="foo", schema="bar", config=config, **kwargs)

    config["handler_type"] = handler_type
    # Validate for all required keys
    for k, v in {**base_config, **db_keys}.items():
        config.pop(k)
        with pytest.raises(ValidationError):
            ds.make_handler_metadata(table="foo", schema="bar", config=config, **kwargs)
        config[k] = v

    handler = ds.make_handler_metadata(
        table="foo", schema="bar", config=config, **kwargs
    )
    assert isinstance(handler, ds.Handler)
    assert isinstance(handler.metadata, expected_type)
    for k, v in {**base_config, **db_keys, **kwargs}.items():
        assert handler.metadata.dict()[k] == v


COLUMNS = [
    ds.DataSourceColumn(
        name="foo", dataType="Integer", remoteType="Integer", nullable=True
    ),
    ds.DataSourceColumn(
        name="bar", dataType="Integer", remoteType="Integer", nullable=False
    ),
    ds.DataSourceColumn(
        name="bat", dataType="Integer", remoteType="Integer", nullable=True
    ),
]

ObjectTestKeys = namedtuple(
    "ObjectTestKeys", ["db_keys", "handler_type", "expected_type", "columns"]
)
OBJECT_TESTS = [
    ObjectTestKeys(
        db_keys={"hostname": "qux"},
        handler_type="PostgreSQL",
        expected_type=ds.PostgresHandlerMetadata,
        columns=COLUMNS,
    ),
    ObjectTestKeys(
        db_keys={"hostname": "qux"},
        handler_type="Redshift",
        expected_type=ds.PostgresHandlerMetadata,
        columns=COLUMNS,
    ),
    ObjectTestKeys(
        db_keys={
            "queryResultLocationBucket": "qux",
            "queryResultLocationDirectory": "quux",
            "region": "quz",
        },
        handler_type="Amazon Athena",
        expected_type=ds.AthenaHandlerMetadata,
        columns=COLUMNS,
    ),
]


@pytest.mark.parametrize("db_keys,handler_type,expected_type,columns", OBJECT_TESTS)
def test_to_immuta_objects(
    db_keys: Dict[str, str],
    handler_type: str,
    expected_type: Any,
    columns: List[ds.DataSourceColumn],
):
    base_config = {"username": "foo", "password": "bar", "database": "baz"}
    config = {**base_config, **db_keys}
    config["handler_type"] = handler_type
    source, handler = ds.to_immuta_objects(
        table="foo", schema="bar", config=config, columns=columns
    )
    assert source.name == ds.make_immuta_table_name(
        table="foo", schema="bar", handler_type=handler_type, user_prefix=""
    )
    assert source.sqlTableName == ds.make_postgres_table_name(
        table="foo", schema="bar", handler_type=handler_type, user_prefix=""
    )
    assert source.blobHandlerType == handler_type
    assert isinstance(handler, ds.Handler)
    assert isinstance(handler.metadata, expected_type)


TABLES = ["foo", "bar", "baz"]
BulkObjectTestKeys = namedtuple(
    "BulkObjectTestKeys", ["db_keys", "handler_type", "expected_type", "tables"]
)
BULK_OBJECT_TESTS = [
    BulkObjectTestKeys(
        db_keys={"hostname": "qux"},
        handler_type="PostgreSQL",
        expected_type=ds.PostgresHandlerMetadata,
        tables=[],
    ),
    BulkObjectTestKeys(
        db_keys={"hostname": "qux"},
        handler_type="PostgreSQL",
        expected_type=ds.PostgresHandlerMetadata,
        tables=TABLES,
    ),
    BulkObjectTestKeys(
        db_keys={"hostname": "qux"},
        handler_type="Redshift",
        expected_type=ds.PostgresHandlerMetadata,
        tables=TABLES,
    ),
    BulkObjectTestKeys(
        db_keys={
            "queryResultLocationBucket": "qux",
            "queryResultLocationDirectory": "quux",
            "region": "quz",
        },
        handler_type="Amazon Athena",
        expected_type=ds.AthenaHandlerMetadata,
        tables=TABLES,
    ),
]


@pytest.mark.parametrize("db_keys,handler_type,expected_type,tables", BULK_OBJECT_TESTS)
def test_make_bulk_create_objects(
    db_keys: Dict[str, str], handler_type: str, expected_type: Any, tables: List[str]
):
    base_config = {"username": "foo", "password": "bar", "database": "baz"}
    config = {**base_config, **db_keys}
    config["handler_type"] = handler_type

    source, handlers = ds.make_bulk_create_objects(
        tables=tables, schema="bar", config=config
    )
    assert len(handlers) == len(tables)
    assert source.blobHandlerType == handler_type

    table_names = []
    for handler in handlers:
        table_names.append(handler.metadata.bodataTableName)
        assert isinstance(handler, ds.Handler)
        assert isinstance(handler.metadata, expected_type)

    for table in tables:
        assert (
            ds.make_immuta_table_name(
                table=table, schema="bar", handler_type=handler_type, user_prefix=""
            )
            in table_names
        )



#!/usr/bin/env python

"""
Script that can be invoked to grep for all existing data sources and then tag columns
within sources as per specified mapping.
"""

import logging

import click
from tqdm import tqdm

from fh_immuta_utils.client import get_client
from fh_immuta_utils.config import parse_config
from fh_immuta_utils.paginator import Paginator
from fh_immuta_utils.tagging import Tagger


@click.command(help="Tag existing data sources based on provided tagging info")
@click.option("--config-file", required=True)
@click.option(
    "--search-text",
    help="Will match all data sources that contains this string anywhere in their name",
)
@click.option(
    "--dry-run",
    is_flag=True,
    default=False,
    help="Log the data stores that would be updated without affecting any change",
)
@click.option("--debug", is_flag=True, default=False, help="Debug logging")
def cli_entrypoint(config_file: str, search_text: str, dry_run: bool, debug: bool):
    return main(
        config_file=config_file, search_text=search_text, dry_run=dry_run, debug=debug
    )


def main(config_file: str, search_text: str, dry_run: bool, debug: bool):
    logging.basicConfig(
        format="[%(name)s][%(levelname)s][%(asctime)s] %(message)s",
        datefmt="%Y-%m-%d %H:%M:%S",
        level=(logging.DEBUG if debug else logging.INFO),
    )
    config = parse_config(config_file=config_file)
    client = get_client(**config)
    tagger = Tagger(config_root=config["config_root"])

    logging.info("Making tags")
    tagger.make_tags(client)

    logging.info("Gathering data sources to tag")
    data_sources_to_tag = []
    with Paginator(client.get_data_source_list, search_text=search_text) as paginator:
        for data_source in paginator:
            data_sources_to_tag.append(
                {"id": data_source["id"], "name": data_source["name"]}
            )

    progress_iterator = tqdm(data_sources_to_tag)
    for data_source in progress_iterator:
        progress_iterator.set_description(
            desc=f"Tagging ID: {data_source['id']}, Name: {data_source['name']} :"
        )
        data_source_tags = tagger.get_tags_for_data_source(name=data_source["name"])
        if data_source_tags:
            logging.debug(f"Adding data source tags to {data_source['name']}.")
            if not dry_run:
                client.tag_data_source(id=data_source["id"], tag_data=data_source_tags)
        dictionary = client.get_data_source_dictionary(id=data_source["id"])
        enriched_columns = tagger.enrich_columns_with_tagging(dictionary.metadata)
        if enriched_columns == dictionary.metadata:
            logging.debug(
                f"No change to column tags for data source: {data_source['name']}. Skipping."
            )
            continue
        logging.debug(
            f"Enriched columns for {data_source['name']}:"
            f" {dictionary.dict()['metadata']}"
        )
        logging.info(
            f"Change detected to column tags. Updating data source {data_source['name']}'s data dictionary."
        )
        dictionary.metadata = enriched_columns
        if not dry_run:
            client.update_data_source_dictionary(
                id=data_source["id"], dictionary=dictionary
            )
    logging.info("FIN.")


if __name__ == "__main__":
    cli_entrypoint()



# Copyright 2018 Immuta, Inc. Licensed under the Immuta Software License
# Version 0.1 (the "License"); you may not use this file except in
# compliance with the License. You may obtain a copy of the License at
#
#    http://www.immuta.com/licenses/Immuta_Software_License_0.1.txt
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import abc
import logging
import os
import warnings
from typing import Dict, Any
import six.moves.urllib_parse as urlparse

import requests
from .exceptions import InvalidTokenError
from .exceptions import UnknownAuthenticationScheme
from .exceptions import ImmutaCredentialsError


class AuthScheme(object):
    __metaclass__ = abc.ABCMeta

    def __init__(self, **kwargs):
        self._session = kwargs.get("session")
        if not self._session:
            self._session = requests.Session()

    @abc.abstractmethod
    def authenticate(self, base_url, ca_certs):
        """Authenticate the user and return an auth token"""
        return ""

    def parse_token_from_response(self, resp_json):
        if not resp_json or "token" not in resp_json:
            raise InvalidTokenError("Unable to obtain token from Immuta")
        return resp_json["token"]

    def handle_response(self, response):
        if 400 <= response.status_code <= 600:
            error_msg = "Error fetching token %d, %s" % (
                response.status_code,
                response.reason,
            )
            raise InvalidTokenError(error_msg)
        return self.parse_token_from_response(response.json())


class UsernamePasswordAuth(AuthScheme):
    AUTH_URL_TEMPLATE = "bim/iam/{iamid}/user/authenticate"

    def __init__(self, iamid, username, password, **kwargs):
        super(UsernamePasswordAuth, self).__init__(**kwargs)
        self.iamid = iamid
        self.username = username
        self.password = password

    def authenticate(self, base_url, ca_certs):
        url = urlparse.urljoin(
            base_url, self.AUTH_URL_TEMPLATE.format(iamid=self.iamid)
        )
        auth_json = {"username": self.username, "password": self.password}
        response = self._session.post(url, json=auth_json, verify=ca_certs)
        return self.handle_response(response)


class ApiKeyAuth(AuthScheme):
    AUTH_URL_TEMPLATE = "bim/apikey/authenticate"

    def __init__(self, apikey, **kwargs):
        super(ApiKeyAuth, self).__init__(**kwargs)
        self.apikey = apikey

    def authenticate(self, base_url, ca_certs):
        url = urlparse.urljoin(base_url, self.AUTH_URL_TEMPLATE)
        response = self._session.post(
            url, json={"apikey": self.apikey}, verify=ca_certs
        )
        return self.handle_response(response)


class OAuth2Auth(AuthScheme):
    AUTH_URL_TEMPLATE = "bim/oauth/token"

    def __init__(self, refresh_token, client_id, client_secret):
        self.refresh_token = refresh_token
        self.client_id = client_id
        self.client_secret = client_secret

    def parse_token_from_response(self, resp_json):
        if not resp_json or "access_token" not in resp_json:
            raise InvalidTokenError("Unable to obtain access token from Immuta")
        return resp_json["access_token"]

    def authenticate(self, base_url, ca_certs):
        url = urlparse.urljoin(base_url, self.AUTH_URL_TEMPLATE)
        response = self._session.post(
            url,
            json={
                "grant_type": "refresh_token",
                "refresh_token": self.refresh_token,
                "client_id": self.client_id,
                "client_secret": self.client_secret,
            },
            verify=ca_certs,
        )
        return self.handle_response(response)


class ImmutaRequestsAuth(requests.auth.AuthBase):
    """Immuta's authentication for the python requests module"""

    def __init__(self, base_url, auth_scheme, ca_certs):
        self.auth_scheme = auth_scheme
        self.token = None
        self.ca_certs = ca_certs
        self.base_url = base_url

    def _add_auth_header_and_retry(self, r, **kwargs):
        self.token = self.auth_scheme.authenticate(self.base_url, self.ca_certs)
        # We want to reuse the original request, so consume content and close the request
        r.content
        r.close()
        prep = r.request.copy()
        prep.headers["Authorization"] = self.__get_auth_string()
        _r = r.connection.send(prep, **kwargs)
        return _r

    def handle_401(self, r, **kwargs):
        if r.status_code == 401:
            return self._add_auth_header_and_retry(r, **kwargs)
        return r

    def handle_500(self, r, **kwargs):
        if self.token is None and r.status_code == 500:
            return self._add_auth_header_and_retry(r, **kwargs)
        return r

    def __call__(self, r):
        if self.token:
            r.headers["Authorization"] = self.__get_auth_string()
        r.register_hook("response", self.handle_401)
        # This is necessary because some of the API endpoints don't handle OAuth correctly.
        # Eg. When posting to /tag without a bearer token, instead of throwing a 401,
        # the API returns 500.
        r.register_hook("response", self.handle_500)
        return r

    def __get_auth_string(self):
        return "Bearer {0}".format(self.token)


def build_auth_scheme(**kwargs):
    """Generates an AuthScheme instance based on the given input.
    When the required set of parameters is found for an AuthScheme implementation, it will be returned.
    The order or precedence is: ApiKeyAuth, UsernamePasswordAuth, and OAuth2Auth.
    If no AuthScheme implementation is found for the given parameters, a UnknownAuthenticationScheme exception is thrown

    Implementation required parameters:
        - ApiKeyAuth:
            - apikey
        - UsernamePasswordAuth:
            - username
            - password
            - iamid
        - OAuth2Auth:
            - refresh_token
            - client_id
            - client_secret

    :param kwargs: dictionary of the named parameters
    :return: An AuthScheme implementation
    """

    if kwargs.get("apiKey"):
        return ApiKeyAuth(kwargs["apiKey"])

    if kwargs.get("username") and kwargs.get("password") and kwargs.get("iamid"):
        return UsernamePasswordAuth(**kwargs)

    if (
        kwargs.get("refresh_token")
        and kwargs.get("client_id")
        and kwargs.get("client_secret")
    ):
        return OAuth2Auth(**kwargs)

    raise UnknownAuthenticationScheme()


def retrieve_credentials_from_vault(credentials_dict: Dict[str, Any]) -> Dict[str, str]:
    """ Expects that you're logged into Hashicorp Vault. """
    import hvac

    client = hvac.Client()
    if not client.is_authenticated():
        raise ImmutaCredentialsError(
            "Vault is not authenticated. Log in to vault first"
        )
    secret = client.secrets.kv.v2.read_secret_version(path=credentials_dict["key"])
    retrieved_credentials = secret["data"]["data"]
    return retrieved_credentials


def retrieve_credentials_from_environment(
    credentials_dict: Dict[str, Any]
) -> Dict[str, str]:
    """ Will throw if env var is unset. """
    try:
        credentials_dict["value"] = os.environ[credentials_dict["key"]]
    except KeyError:
        raise ImmutaCredentialsError(f"No value for env var {credentials_dict['key']}")
    return credentials_dict


def retrieve_credentials_from_local_dict(
    credentials_dict: Dict[str, Any]
) -> Dict[str, str]:
    return credentials_dict


def retrieve_credentials(credentials_dict: Dict[str, Any]) -> Dict[str, str]:
    func = CREDENTIALS_RESOLVERS[credentials_dict["source"]]
    return func(credentials_dict)


CREDENTIALS_RESOLVERS = {
    "VAULT": retrieve_credentials_from_vault,
    "ENV": retrieve_credentials_from_environment,
    "LOCAL": retrieve_credentials_from_local_dict,
}



import os
from typing import Dict, Any, Optional, List, Set
from enum import Enum, unique
import logging
import glob

import yaml
from pydantic import BaseModel, Field

from .tagging import Tagger


@unique
class CircumstanceType(Enum):
    TAG = "tags"
    COLUMN_TAG = "columnTags"


@unique
class ActionType(Enum):
    MASKING = "masking"


class PolicyGroup(BaseModel):
    name: str
    iam: str


class ColumnTag(BaseModel):
    name: str
    # Set to True if it's a root tag
    hasLeafNodes: bool


class DataSourceTag(BaseModel):
    name: str
    # Set to True if it's a root tag
    hasLeafNodes: bool


class PolicyAuthorization(BaseModel):
    auth: str
    value: str
    iam: str


class PolicyCondition(BaseModel):
    type: str


class GroupCondition(PolicyCondition):
    type: str = "groups"
    group: PolicyGroup
    field: Optional[str] = ""

    def dict(
        self,
        *,
        include: Set[str] = None,
        exclude: Set[str] = None,
        by_alias: bool = False,
        exclude_unset: bool = False,
        exclude_defaults: bool = False,
        exclude_none: bool = False,
    ) -> Dict[str, Any]:
        """
        The 'field' key should not exist in the payload if it has no value
        as Immuta's API will reject the request. This is because group condition
        can be created in different contexts, some of which require setting 'field'
        and others where 'field' makes no sense.
        """
        if self.field == "":
            exclude = {"field"}
        return super().dict(
            include=include,
            exclude=exclude,
            by_alias=by_alias,
            exclude_unset=exclude_unset,
            exclude_defaults=exclude_defaults,
            exclude_none=exclude_none,
        )


class AuthorizationCondition(PolicyCondition):
    type: str = "authorizations"
    authorization: PolicyAuthorization
    field: str


class PurposeCondition(PolicyCondition):
    type: str = "purposes"
    value: str
    field: str


class PolicyCircumstance(BaseModel):
    operator: str
    type: str


class ColumnTagCircumstance(PolicyCircumstance):
    type: str = "columnTags"
    columnTag: ColumnTag


class TagCircumstance(PolicyCircumstance):
    type: str = "tags"
    tag: DataSourceTag


class PolicyExceptions(BaseModel):
    # Must be one of 'and', 'or'
    operator: str
    conditions: List[PolicyCondition]


class PolicyRuleConfig(BaseModel):
    policy_fields: Optional[List[ColumnTag]] = Field(required=False, alias="fields")


class MaskingConfig(BaseModel):
    type: str
    # Unsure what this is
    metadata: Optional[Dict] = {}


class MaskingRuleConfig(PolicyRuleConfig):
    maskingConfig: MaskingConfig


class PolicyRule(BaseModel):
    type: str
    # Policies will not necessarily have any exceptions ("apply to everyone")
    exceptions: Optional[PolicyExceptions]
    config: PolicyRuleConfig


class PolicyAction(BaseModel):
    type: str


class SubscriptionPolicyAction(BaseModel):
    type: str = "subscription"
    subscriptionType: str
    exceptions: Optional[PolicyExceptions] = None
    allowDiscovery: bool = False


class DataPolicyAction(PolicyAction):
    rules: Optional[List[PolicyRule]] = None
    description: Optional[str] = ""


class MaskingAction(DataPolicyAction):
    type: str = "masking"


class GlobalPolicy(BaseModel):
    name: str
    id: Optional[int] = None
    type: str
    template: bool = False
    # Don't know what this is supposed to be.
    # No mention in the docs
    ownerRestrictions: Optional[Any] = None
    circumstances: List[PolicyCircumstance]

    def dict(
        self,
        *,
        include: Set[str] = None,
        exclude: Set[str] = None,
        by_alias: bool = False,
        exclude_unset: bool = False,
        exclude_defaults: bool = False,
        exclude_none: bool = False,
    ) -> Dict[str, Any]:
        # Remove keys that should not exist in the payload if they have no values
        exclude = set()
        if not self.id:
            exclude.add("id")
        return super().dict(
            include=include,
            exclude=exclude,
            by_alias=by_alias,
            exclude_unset=exclude_unset,
            exclude_defaults=exclude_defaults,
            exclude_none=exclude_none,
        )


class GlobalDataPolicy(GlobalPolicy):
    type: str = "data"
    actions: List[Dict]


class GlobalSubscriptionPolicy(GlobalPolicy):
    type: str = "subscription"
    staged: bool = True
    actions: List[Dict]


class PolicyConfig:
    """
    Wrapper around managing data & subscription policy configuration
    """

    def __init__(self, config_root: str) -> None:
        self.data_policy_config: Dict[str, Any] = {}
        self.subscription_policy_config: Dict[str, Any] = {}
        self.read_configs(config_root=config_root)

    def _read_data_configs(self, config_root: str) -> None:
        for data_policy_file in glob.glob(
            os.path.join(config_root, "policies/data", "*.yml")
        ):
            logging.debug("Reading data policy file: %s", data_policy_file)
            with open(data_policy_file) as handle:
                contents = yaml.safe_load(handle)
                self.data_policy_config = {
                    **self.data_policy_config,
                    **contents.get("DATA_POLICIES", {}),
                }

    def _read_subscription_configs(self, config_root: str) -> None:
        for subscription_policy_file in glob.glob(
            os.path.join(config_root, "policies/subscription", "*.yml")
        ):
            logging.debug(
                "Reading subscription policy file: %s", subscription_policy_file
            )
            with open(subscription_policy_file) as handle:
                contents = yaml.safe_load(handle)
                self.subscription_policy_config = {
                    **self.subscription_policy_config,
                    **contents.get("SUBSCRIPTION_POLICIES", {}),
                }

    def read_configs(self, config_root: str) -> None:
        self._read_data_configs(config_root=config_root)
        self._read_subscription_configs(config_root=config_root)


def make_policy_exceptions(
    iam_groups: List[str], operator: str = "or"
) -> PolicyExceptions:
    conditions = []
    for group in iam_groups:
        # TODO: Generalize for other IAMs and conditions
        conditions.append(GroupCondition(group=PolicyGroup(name=group, iam="okta")))
    return PolicyExceptions(operator=operator, conditions=conditions)


def make_policy_circumstance(
    tag: str, tagger: Tagger, circumstance_type: Any, operator: str = "or"
) -> Any:
    try:
        circumstance_type = CircumstanceType(circumstance_type)
    except ValueError:
        raise

    if circumstance_type is CircumstanceType.TAG:
        return TagCircumstance(
            operator=operator,
            tag=DataSourceTag(name=tag, hasLeafNodes=tagger.is_root_tag(tag)),
        )
    elif circumstance_type is CircumstanceType.COLUMN_TAG:
        return ColumnTagCircumstance(
            operator=operator,
            columnTag=ColumnTag(name=tag, hasLeafNodes=tagger.is_root_tag(tag)),
        )


def make_policy_object_from_json(json_policy: Dict[str, Any]) -> GlobalPolicy:
    circumstances = []
    exceptions: Optional[PolicyExceptions] = None
    if json_policy["circumstances"] is not None:
        for circumstance in json_policy["circumstances"]:
            if circumstance["type"] == "columnTags":
                circumstances.append(ColumnTagCircumstance(**circumstance))
            elif circumstance["type"] == "tags":
                circumstances.append(TagCircumstance(**circumstance))
            else:
                raise TypeError(
                    f"Unsupported type for policy circumstance: {circumstance['type']}"
                )

    actions = []
    for action in json_policy["actions"]:
        if action.get("exceptions"):
            conditions = []
            for condition in action["exceptions"]["conditions"]:
                if condition["type"] == "groups":
                    conditions.append(GroupCondition(**condition))
                elif condition["type"] == "authorizations":
                    conditions.append(AuthorizationCondition(**condition))
                else:
                    raise TypeError(
                        f"Unsupported type for policy condition: {condition['type']}"
                    )
            exceptions = PolicyExceptions(
                operator=action["exceptions"]["operator"], conditions=conditions
            )
        else:
            exceptions = None
        actions.append(action)

    if json_policy["type"] == "subscription":
        return GlobalSubscriptionPolicy(
            id=json_policy["id"],
            name=json_policy["name"],
            type=json_policy["type"],
            template=json_policy["template"],
            circumstances=circumstances,
            actions=actions,
        )
    elif json_policy["type"] == "data":
        return GlobalDataPolicy(
            id=json_policy["id"],
            name=json_policy["name"],
            type=json_policy["type"],
            template=json_policy["template"],
            circumstances=circumstances,
            actions=actions,
        )
    else:
        raise TypeError(f"Unsupported type for Global policy: {json_policy['type']}")


def make_subscription_policy_action(
    exceptions_config: Dict, allow_discovery: bool, tagger: Tagger
) -> SubscriptionPolicyAction:

    iam_groups = []
    for condition in exceptions_config["conditions"]:
        iam_groups.extend(condition["iam_groups"])
    return SubscriptionPolicyAction(
        type="subscription",
        subscriptionType="policy",
        allowDiscovery=allow_discovery,
        exceptions=make_policy_exceptions(
            iam_groups=iam_groups, operator=exceptions_config["operator"]
        ),
    )


def make_global_subscription_policy(
    policy_name: str, policy_config: Dict, tagger: Tagger
) -> GlobalSubscriptionPolicy:

    actions: List[SubscriptionPolicyAction] = []

    if policy_config.get("actions"):
        for action_grouping in policy_config["actions"]:
            action = make_subscription_policy_action(
                exceptions_config=action_grouping["exceptions"],
                allow_discovery=action_grouping.get("allowDiscovery", False),
                tagger=tagger,
            )
            actions.append(action)
    else:
        raise KeyError(f"Missing actions for subscription policy: {policy_name}")

    circumstances: List[Any] = []
    if policy_config.get("circumstances"):
        for circumstance_grouping in policy_config["circumstances"]:
            for tag in circumstance_grouping["tags"]:
                circumstance = make_policy_circumstance(
                    tag=tag,
                    tagger=tagger,
                    circumstance_type=circumstance_grouping["type"],
                    operator=circumstance_grouping["operator"],
                )
                circumstances.append(circumstance)
    else:
        raise KeyError(f"Missing circumstances for subscription policy: {policy_name}")

    return GlobalSubscriptionPolicy(
        name=policy_name,
        circumstances=circumstances,
        actions=actions,
        staged=policy_config.get("staged", True),
    )


def make_policy_rule(
    rule_type: str,
    exceptions_config: Dict,
    config_field_tags: List[str],
    tagger: Tagger,
) -> PolicyRule:
    config_rule_fields: List[ColumnTag] = []
    # TODO: handle other types of config
    for tag in config_field_tags:
        config_rule_fields.append(
            ColumnTag(name=tag, hasLeafNodes=tagger.is_root_tag(tag))
        )

    # TODO: pass in conditions instead of iam_groups to make_policy_exceptions after subscription policy refactor
    iam_groups = []
    for condition in exceptions_config["conditions"]:
        iam_groups.extend(condition["iam_groups"])
    return PolicyRule(
        type=rule_type,
        exceptions=make_policy_exceptions(
            iam_groups=iam_groups,
            operator=exceptions_config["operator"],
        ),
        # TODO: Support other config types
        config=MaskingRuleConfig(
            fields=config_rule_fields,
            maskingConfig=MaskingConfig(type="Consistent Value"),
        ),
    )


def make_data_policy_action(
    action_type: Any, rules_config: Dict, tagger: Tagger
) -> MaskingAction:
    try:
        action_type = ActionType(action_type)
    except ValueError:
        raise

    rules = []
    for rule in rules_config:
        rules.append(
            make_policy_rule(
                rule_type=rule["type"],
                exceptions_config=rule["exceptions"],
                config_field_tags=rule["config"]["fields"]["tags"],
                tagger=tagger,
            )
        )

    if action_type is ActionType.MASKING:
        return MaskingAction(
            type=ActionType.MASKING.value,
            rules=rules,
        )
    else:
        raise TypeError(f"Unsupported type for action: {action_type}")


def make_global_data_policy(
    policy_name: str, policy_config: Dict, tagger: Tagger
) -> GlobalDataPolicy:
    """
    Returns a GlobalDataPolicy object containing lists of actions and circumstances.
    Actions define what the policy restricts, how it restricts, and for whom it restricts.
    Circumstances define where and how the policy is applied to data sources in Immuta.
    """
    actions: List[MaskingAction] = []
    circumstances: List[Any] = []

    if policy_config.get("actions"):
        for action_grouping in policy_config["actions"]:
            action = make_data_policy_action(
                action_type=action_grouping["type"],
                rules_config=action_grouping["rules"],
                tagger=tagger,
            )
            actions.append(action)
    else:
        raise KeyError(f"Missing actions for data policy: {policy_name}")

    if policy_config.get("circumstances"):
        for circumstance_grouping in policy_config["circumstances"]:
            for tag in circumstance_grouping["tags"]:
                circumstance = make_policy_circumstance(
                    tag=tag,
                    tagger=tagger,
                    circumstance_type=circumstance_grouping["type"],
                    operator=circumstance_grouping["operator"],
                )
                circumstances.append(circumstance)
    else:
        raise KeyError(f"Missing circumstances for data policy: {policy_name}")

    return GlobalDataPolicy(
        name=policy_name,
        circumstances=circumstances,
        actions=actions,
    )



"""Global cli combining all of the commands into one larger piece

"""

import click

from .bulk_delete_data_source import main as bulk_delete_data_source_command
from .manage_data_sources import cli_entrypoint as manage_data_sources_command
from .manage_policies import cli_entrypoint as manage_policies_command
from .tag_existing_data_sources import (
    cli_entrypoint as tag_existing_data_sources_command,
)

main_cli = click.Group()
data_source_cli = click.Group(
    "data-source", help="Manage data sources along with their tagging"
)
data_source_cli.add_command(bulk_delete_data_source_command, "bulk-delete")
data_source_cli.add_command(manage_data_sources_command, "manage")
data_source_cli.add_command(tag_existing_data_sources_command, "tag-existing")
main_cli.add_command(data_source_cli, "data-source")
main_cli.add_command(manage_policies_command, "policies")

if __name__ == "__main__":
    main_cli()



import logging

import click
from tqdm import tqdm

from fh_immuta_utils.client import get_client
from fh_immuta_utils.config import parse_config
from fh_immuta_utils.paginator import Paginator

LOGGER = logging.getLogger(__name__)


@click.command(help="Bulk delete data sources that match a given prefix")
@click.option("--config-file", required=True)
@click.option(
    "--search-text",
    help="Delete the data sources that contains this string in their name",
)
@click.option(
    "--hard-delete",
    is_flag=True,
    default=False,
    help="Permanently remove the data-sources",
)
@click.option(
    "--dry-run",
    is_flag=True,
    default=False,
    help="Log the data stores that would be removed instead of deleting them",
)
@click.option("--debug", is_flag=True, default=False, help="Debug logging")
def main(
    config_file: str, search_text: str, hard_delete: bool, dry_run: bool, debug: bool
):
    logging.basicConfig(
        format="[%(name)s][%(levelname)s][%(asctime)s] %(message)s",
        datefmt="%Y-%m-%d %H:%M:%S",
        level=(logging.DEBUG if debug else logging.INFO),
    )
    config = parse_config(config_file=config_file)
    client = get_client(**config)

    logging.info("Gathering data-stores to delete")
    data_sources_to_delete = []
    with Paginator(client.get_data_source_list, search_text=search_text) as paginator:
        for data_source in paginator:
            data_sources_to_delete.append(
                {"id": data_source["id"], "name": data_source["name"]}
            )

    if dry_run:
        logging.info("bulk-delete dry run")
        for data_source in data_sources_to_delete:
            logging.info(
                f"Data source Id: {data_source['id']}. Name: {data_source['name']}"
            )
    elif hard_delete:
        logging.info(
            f"Hard deleting {len(data_sources_to_delete)} data sources. "
            "The data sources will not be able to be restored in the future"
        )
        for data_source in tqdm(data_sources_to_delete, desc="Deleting"):
            logging.debug(f"Hard deleting {data_source['name']}")
            client.delete_data_source(data_source["id"])
    else:
        logging.info(
            f"Disabling {len(data_sources_to_delete)} data sources. "
            "The data sources can be restored in the future"
        )
        for data_source in tqdm(data_sources_to_delete, desc="Disabling"):
            logging.debug(f"Disabling {data_source['name']}")
            client.disable_data_source(data_source["id"])


if __name__ == "__main__":
    main()



""" Provides methods to read immuta-utils config files """
import os
from typing import Dict, Any

import yaml

from .exceptions import BadImmutaConfigException

REQUIRED_KEYS = ["base_url", "config_root", "auth_config"]


def parse_config(config_file: str) -> Dict[str, Any]:
    """ Validates the given config file and returns as dict """
    with open(config_file) as handle:
        config = yaml.safe_load(handle)
    for key in REQUIRED_KEYS:
        if key not in config:
            raise BadImmutaConfigException(
                f"Must specify value for {key} in config file {config_file}"
            )
    # If config_root is relative, replace with absolute path
    if not os.path.isabs(config["config_root"]):
        config["config_root"] = os.path.abspath(
            os.path.join(os.path.dirname(config_file), config["config_root"])
        )
    config["config_root"] = os.path.abspath(config["config_root"])
    auth_config = config["auth_config"]
    if auth_config["scheme"] == "ApiKeyAuth":
        if "apiKey" not in auth_config:
            raise BadImmutaConfigException(
                "apiKey required when using key-based auth scheme"
            )
    if auth_config["scheme"] == "UsernamePasswordAuth":
        for key in ["username", "password", "iamid"]:
            if key not in auth_config:
                raise BadImmutaConfigException(
                    f"Must specify value for {key} when using UsernamePasswordAuth"
                )
    if auth_config["scheme"] == "OAuth2Auth":
        for key in ["refresh_token", "client_id", "client_secret"]:
            if key not in auth_config:
                raise BadImmutaConfigException(
                    f"Must specify value for {key} when using OAuth2Auth"
                )
    return config



import os
from unittest import mock

import pytest

import fh_immuta_utils.policy as pol
import fh_immuta_utils.tagging as tg


@pytest.fixture
def tagger():
    with mock.patch("fh_immuta_utils.tagging.Tagger.read_configs", return_value=None):
        obj = tg.Tagger(config_root="")
        obj.tag_map_datadict = {}
        obj.tag_map_datasource = {}
    return obj


@pytest.fixture
def policy_dict():
    return pol.PolicyConfig(
        config_root=f"{os.path.dirname(os.path.abspath(__file__))}/fixtures"
    )


@pytest.fixture
def data_policy_actions_dict(policy_dict):
    return policy_dict.data_policy_config["correct_policy"]["actions"]


@pytest.fixture
def data_policy_circumstances_dict(policy_dict):
    return policy_dict.data_policy_config["correct_policy"]["circumstances"]


@pytest.fixture
def subscription_policy_actions_dict(policy_dict):
    return policy_dict.subscription_policy_config["valid_db_sub_policy"]["actions"]


@pytest.fixture
def subscription_policy_staged_bool(policy_dict):
    return policy_dict.subscription_policy_config["valid_db_sub_policy"]["staged"]


def test_make_policy_exceptions(data_policy_actions_dict):
    conditions = data_policy_actions_dict[0]["rules"][0]["exceptions"]["conditions"]
    iam_groups = conditions[0]["iam_groups"]
    expected_exceptions = pol.PolicyExceptions(
        operator="and",
        conditions=[
            pol.GroupCondition(group=pol.PolicyGroup(name="group1", iam="okta"))
        ],
    )
    exceptions = pol.make_policy_exceptions(iam_groups=iam_groups, operator="and")
    assert exceptions == expected_exceptions


def test_make_policy_circumstance(data_policy_circumstances_dict, tagger):
    circumstance_type = data_policy_circumstances_dict[0]["type"]
    tags = data_policy_circumstances_dict[0]["tags"]
    expected_circumstance = pol.ColumnTagCircumstance(
        operator="or",
        columnTag=pol.ColumnTag(name=tags[0], hasLeafNodes=tagger.is_root_tag(tags[0])),
    )
    circumstance = pol.make_policy_circumstance(
        tag=tags[0],
        tagger=tagger,
        circumstance_type=circumstance_type,
    )
    assert circumstance == expected_circumstance


def test_make_policy_circumstance_bad_type(data_policy_circumstances_dict, tagger):
    circumstance_type = "foo"
    tags = data_policy_circumstances_dict[0]["tags"]
    with pytest.raises(ValueError):
        return pol.make_policy_circumstance(
            tag=tags[0],
            tagger=tagger,
            circumstance_type=circumstance_type,
        )


def test_make_policy_rule(data_policy_actions_dict, tagger):
    example_rule = data_policy_actions_dict[0]["rules"][0]
    rule_type = example_rule["type"]
    config_field_tags = example_rule["config"]["fields"]["tags"]
    exceptions_config = example_rule["exceptions"]
    iam_groups = exceptions_config["conditions"][0]["iam_groups"]
    operator = exceptions_config["operator"]

    expected_rule = pol.PolicyRule(
        type=rule_type,
        exceptions=pol.make_policy_exceptions(iam_groups=iam_groups, operator=operator),
        config=pol.MaskingRuleConfig(
            fields=[
                pol.ColumnTag(
                    name=config_field_tags[0],
                    hasLeafNodes=tagger.is_root_tag(config_field_tags[0]),
                )
            ],
            maskingConfig=pol.MaskingConfig(type="Consistent Value"),
        ),
    )
    rule = pol.make_policy_rule(
        rule_type=rule_type,
        exceptions_config=exceptions_config,
        config_field_tags=config_field_tags,
        tagger=tagger,
    )
    assert rule == expected_rule


def test_make_data_policy_action(data_policy_actions_dict, tagger):
    action_type = data_policy_actions_dict[0]["type"]
    rules_config = data_policy_actions_dict[0]["rules"]
    rule1 = rules_config[0]
    rule1_conditions = rule1["exceptions"]["conditions"]
    rule1_iam_groups = rule1_conditions[0]["iam_groups"]
    rule2 = rules_config[1]
    rule2_conditions = rule2["exceptions"]["conditions"]
    rule2_iam_groups = (
        rule2_conditions[0]["iam_groups"] + rule2_conditions[1]["iam_groups"]
    )
    expected_action = pol.MaskingAction(
        type=action_type,
        rules=[
            pol.PolicyRule(
                type=rule1["type"],
                exceptions=pol.make_policy_exceptions(
                    iam_groups=rule1_iam_groups,
                    operator=rule1["exceptions"]["operator"],
                ),
                config=pol.MaskingRuleConfig(
                    fields=[
                        pol.ColumnTag(
                            name=rule1["config"]["fields"]["tags"][0],
                            hasLeafNodes=tagger.is_root_tag(
                                rule1["config"]["fields"]["tags"][0]
                            ),
                        )
                    ],
                    maskingConfig=pol.MaskingConfig(type="Consistent Value"),
                ),
            ),
            pol.PolicyRule(
                type=rule2["type"],
                exceptions=pol.make_policy_exceptions(
                    iam_groups=rule2_iam_groups,
                    operator=rule2["exceptions"]["operator"],
                ),
                config=pol.MaskingRuleConfig(
                    fields=[
                        pol.ColumnTag(
                            name=rule2["config"]["fields"]["tags"][0],
                            hasLeafNodes=tagger.is_root_tag(
                                rule2["config"]["fields"]["tags"][0]
                            ),
                        ),
                        pol.ColumnTag(
                            name=rule2["config"]["fields"]["tags"][1],
                            hasLeafNodes=tagger.is_root_tag(
                                rule2["config"]["fields"]["tags"][1]
                            ),
                        ),
                    ],
                    maskingConfig=pol.MaskingConfig(type="Consistent Value"),
                ),
            ),
        ],
    )
    action = pol.make_data_policy_action(
        action_type=action_type,
        rules_config=rules_config,
        tagger=tagger,
    )
    assert action == expected_action


def test_make_data_policy_action_bad_type(data_policy_actions_dict, tagger):
    action_type = "foo"
    rules_config = data_policy_actions_dict[0]["rules"]
    with pytest.raises(ValueError):
        return pol.make_data_policy_action(
            action_type=action_type,
            rules_config=rules_config,
            tagger=tagger,
        )


def test_make_subscription_policy_action(subscription_policy_actions_dict, tagger):
    exceptions_config = subscription_policy_actions_dict[0]["exceptions"]
    exception_conditions = exceptions_config["conditions"]
    iam_groups = exception_conditions[0]["iam_groups"]
    expected_action = pol.SubscriptionPolicyAction(
        type="subscription",
        subscriptionType="policy",
        allowDiscovery=False,
        exceptions=pol.make_policy_exceptions(
            iam_groups=iam_groups, operator=exceptions_config["operator"]
        ),
    )
    action = pol.make_subscription_policy_action(
        exceptions_config, subscription_policy_actions_dict[0]["allowDiscovery"], tagger
    )
    assert action == expected_action


def test_subscription_policy_staged(subscription_policy_staged_bool):
    assert subscription_policy_staged_bool == False



# coding: utf-8

from setuptools import setup, find_packages  # noqa: H301

NAME = "fh-immuta-utils"


with open("README.md", "r") as fh:
    long_description = fh.read()


setup(
    name=NAME,
    version="0.2.0",
    entry_points={
        "console_scripts": ["fh-immuta-utils = fh_immuta_utils.scripts.cli:main_cli"]
    },
    description="Flatiron Immuta API",
    author_email="data-tooling@flatiron.com",
    url="https://github.com/flatironhealth/fh-immuta-utils",
    packages=find_packages(),
    include_package_data=True,
    install_requires=[
        "click",
        "hvac",
        "pydantic",
        "PyYAML",
        "requests",
        "six",
        "toolz",
        "tqdm",
        "urllib3",
    ],
    long_description=long_description,
    long_description_content_type="text/markdown",
    classifiers=[
        "Programming Language :: Python :: 3 :: Only",
        "License :: OSI Approved :: BSD License",
        "Operating System :: OS Independent",
        "Development Status :: 5 - Production/Stable",
    ],
    python_requires=">=3.6",
)
